// *******************************************************************
//    IBM Rational Ltd  Rhapsody in C++ Simplified eXecution Framework
//   ================== ==============================================
//
//    Model Element   : OMTimeoutPool
//    File Path       : OMTimeoutPool.cpp 
//    Licensed Materials - Property of IBM
//    (c) Copyright IBM Corporation 2008, 2016. All Rights Reserved.
//    
//    Note to U.S. Government Users Restricted Rights:  
//    Use, duplication or disclosure restricted by GSA ADP Schedule 
//    Contract with IBM Corp.
// *******************************************************************


#include "OMTimeoutPool.h"
#include "OMMainThread.h"
#include "OMReactive.h"
//## package SXF::OMGeneric_Pkg

//## class OMTimeoutPool
//#[ requirement(s)
// Realizes requirement TimeoutPool Class #SXF_LR.1.8:
// This class shall manage all the timeouts, keeping two lists, one free list and one busy list. Both lists shall contain elements of static array, which shall be defined in main function of application.
// As timeouts are allocated, they move from the free list to the busy list.
// When the timeouts are unscheduled or expire and processed, they are moved back from the busy list to the free list.
// Timeouts are created initially and not deleted.
//#]
//#[ requirement(s)
// Realizes requirement Timeout Pool Constructor #SXF_LR.1.8.6:
// The OMTimeoutPool constructor shall create its 'tmPoolGuard' timeout pool guard  and store system time in the lastTime attribute.
// It initializes its busyList to be an empty list.
// It initializes its freeList to be an empty list.
// The constructor shall be public.
//#]
OMTimeoutPool::OMTimeoutPool(void) : lastTime(OMOSSpecific::getSystemTime()), itsBusyList(NULL), itsFreeList(NULL) {
    //#[ operation OMTimeoutPool()
    (void)OMOSSpecific::createOSMutex(tmPoolGuard);
    //#]
}

//#[ requirement(s)
// Realizes requirement Add Timeout to Busy List #SXF_LR.1.8.4:
// OMTimeoutPool::addTm function shall add timeout into (sorted) list of busy elements.
// Function has exactly two parameters:
// - the first parameter (called 'delayTime') has type 'const OMOSSpecific::timeUnit'.
// - the second parameter (called 'destination') has type 'OMReactive* const'.
// Function returns a value of type 'OMTimeout*': If the 'itsFreeList'-field is NULL, addTm returns NULL. 
// Otherwise the function returns the address of the OMTimeout referenced by itsFreeList.
// The new timeout with same delivery time as the existing timeout shall be added after the first such timeout. 
// All actions shall be done under timeout pool guard protection.
// The function shall be public.
// Realizes requirement OMTimeoutPool_addTm_effect_deleteAfterConsume #SXF_LR.1.8.4.1:
// If itsFreeList is not NULL, let itsFreeList (at the time addTm is called) be the "previously free Timeout" in the following. One postcondition of OMTimeoutPool::addTm is that the deleteAfterConsume-flag of the previously free Timeout is equal 'false'.
// Realizes requirement OMTimeoutPool_addTm_effect_deliveryTime #SXF_LR.1.8.4.2:
// If itsFreeList is not NULL, let itsFreeList (at the time addTm is called) be the "previously free Timeout" in the following. One postcondition of OMTimeoutPool::addTm is that the delivery time of the previously free Timeout is equal to the sum of the delayTime parameter value and the current OS-system time.
// Realizes requirement OMTimeoutPool_addTm_effect_destination #SXF_LR.1.8.4.3:
// If itsFreeList is not NULL, let itsFreeList (at the time addTm is called) be the "previously free Timeout" in the following. One postcondition of OMTimeoutPool::addTm is that the destination of the previously free Timeout is equal to the 'destination'-parameter value of addTm.
// Realizes requirement OMTimeoutPool_addTm_effect_itsBusyList #SXF_LR.1.8.4.4:
// If itsFreeList is NULL, let itsFreeList (at the time addTm is called) be the "previously free Timeout" in the following, and itsBusyList (at the time addTm is called) be the "previous busy list". 
// One postcondition of OMTimeoutPool::addTm is that the previously free Timeout is sorted into the previous busy list, 
// ordered by the deliveryTime of the timeouts in the list (starting from small to large), and taking the following special order rule into account for equal delivery times:
// The previously free timeout is sorted into the previous busy list directly after the first timeout with the same delivery Time.
// The end of the busy list is denoted as a NULL, that is the empty list is denoted by itsBusyList==NULL.
// Realizes requirement OMTimeoutPool_addTm_effect_itsFreeList #SXF_LR.1.8.4.5:
// If itsFreeList is NULL, OMTimeoutPool::addTm does not change itsFreeList. If itsFreeList is not NULL, let itsFreeList->itsNextTimeout denote the "next free Timeout" in the following, and itsFreeList the "previously free Timeout". One postcondition of OMTimeoutPool::addTm is that itsFreeList is equal to the next free Timeout, and the previously free Timeout is not part of the itsFreeList any more.
//#]
/*LDRA_INSPECTED 120 D 2 : MISRA-C++:2008 7-1-2: "destination" parameter cannot be const */
OMTimeout* OMTimeoutPool::addTm(const OMOSSpecific::timeUnit delayTime, OMReactive* const destination) {
    //#[ operation addTm(timeUnit,OMReactive)
    (void)OMOSSpecific::lock(tmPoolGuard);
    OMTimeout* aTimeout = NULL;
    // Do we have a timeout available ?
    if ( itsFreeList != NULL )  
    { 
    	OMOSSpecific::timeUnit dueTime;
    	// Get a timeout from the free list
    	aTimeout = itsFreeList;
    	itsFreeList = aTimeout->getItsNextTimeout(); 
    	
    	  // Initialize  aTimeout with appropriate info  
    	dueTime = delayTime  + OMOSSpecific::getSystemTime();
    	aTimeout->setDestination( destination );
    	aTimeout->setDeliveryTime( dueTime );
    	aTimeout->setDeleteAfterConsume(false);
    	
    	// and insert timeout onto empty busy list
    	if (itsBusyList == NULL)
    	{
    		aTimeout->setItsNextTimeout( itsBusyList );
    		itsBusyList = aTimeout;
    	}
    	else 
    	{   // insert timeout according to its expiration time 
    		/*LDRA_INSPECTED 70 D : MISRA-C++:2008 0-1-6,0-1-9: "next" is used in the loop below */
    		OMTimeout* next = itsBusyList->getItsNextTimeout();
    		OMTimeout* current =  itsBusyList;
    		OMTimeout* previous = NULL;
    		/*LDRA_INSPECTED 69 D : MISRA-C++:2008 8-5-1: LDRA analysis appears incorrect,there are no uninitialized variables inside of loop */
    		while (current != NULL) 
    		{
    			if (current->getDeliveryTime() > dueTime)
    			{
    				// new timeout expired before the current and exit from loop
    				aTimeout->setItsNextTimeout(current);
    				if (previous == NULL) 
    				{   // put it in the beginning of busy list
    					itsBusyList = aTimeout;
    				}
    				else
    				{   // put it after the previous 
    					previous->setItsNextTimeout(aTimeout);
    				}
    				current = NULL;
    			}
    			else
    			if (next == NULL)
    			{
    				// put it after the current and exit from loop
    				aTimeout->setItsNextTimeout(NULL);
    				current->setItsNextTimeout(aTimeout);
    				current = NULL;				
    			}
    			else 
    			{   // continue to search for tm with later expiration  
    				/*LDRA_INSPECTED 70 D : MISRA-C++:2008 0-1-6,0-1-9: "previous" is used in the loop */
    				previous = current;
    				/*LDRA_INSPECTED 70 D : MISRA-C++:2008 0-1-6,0-1-9: "next" is used in the loop */
    				current = next;
    				next = next->getItsNextTimeout();
    			} 
    		}
    	}
    }
    (void)OMOSSpecific::unlock(tmPoolGuard);
    /*LDRA_INSPECTED 71 S : MISRA-C++:2008 7-5-2: Getting back a pointer to an element in static timeout pool */
    return  aTimeout; 
    
    //#]
}

//#[ requirement(s)
// Realizes requirement Get Expired Timeout #SXF_LR.1.8.5:
// This function shall take the first expired timeout, remove it from list of busy elements and return it. 
// The function returns NULL, if there are no expired timeouts in the list or the list is empty.
// Return type is 'OMTimeout*'.
// All actions shall be done under timeout pool guard protection.
// The function shall be public.
//#]
OMTimeout* OMTimeoutPool::getExpiredTimeout(void) {
    //#[ operation getExpiredTimeout()
    (void)OMOSSpecific::lock(tmPoolGuard);
    OMTimeout* aTimeout = itsBusyList; 
    const OMOSSpecific::timeUnit CurrentTime = OMOSSpecific::getSystemTime();
    
    if ((aTimeout != NULL) && (aTimeout->isExpired(CurrentTime,lastTime)) ) 
    {
    	// remove from the busy list
    	itsBusyList = aTimeout->getItsNextTimeout();
    	// But don't add it to the free list until it has been processed!
    	// return the expired timeout
    }
    else
    {
    	/*LDRA_INSPECTED 69 D : MISRA-C++:2008 8-5-1: LDRA analysis appears incorrect, variable aTimeout is initialized before */
    	aTimeout = NULL;
    }                   
    lastTime = CurrentTime;
    (void)OMOSSpecific::unlock(tmPoolGuard);
    /*LDRA_INSPECTED 71 S : MISRA-C++:2008 7-5-2: Getting back a pointer to an element in static timeout pool */
    return aTimeout;
    //#]
}

//#[ requirement(s)
// Realizes requirement Put Timeout to Free List #SXF_LR.1.8.8:
// OMTimeoutPool::putBack function shall add timeout as the first element into list of free elements.
// The function has one parameter of type 'OMTimeout* const'.
// The element, which is put to the 'itsFreeList' by OMTimeoutPool::pubBack last, is the first removed from the list by OMTimeoutPool::addTm (lifo-order, like a stack).
// It shall be done under timeout pool guard protection.
// The function shall be public.
// 
//#]
void OMTimeoutPool::putBack(OMTimeout* const aTimeout) {
    //#[ operation putBack(OMTimeout)
    if (aTimeout != NULL)
    {
    	(void)OMOSSpecific::lock(tmPoolGuard);
    	aTimeout->setItsNextTimeout( itsFreeList );
    	itsFreeList = aTimeout;
    	(void)OMOSSpecific::unlock(tmPoolGuard);
    }
    
    //#]
}

//#[ requirement(s)
// Realizes requirement Remove Timeout from Busy List #SXF_LR.1.8.3:
// This function shall find timeout in list of busy elements, remove it and put into the list of free elements. 
// Both link lists shall be modified appropriately.
// All actions shall be done under timeout pool guard ('tmPoolGuard') protection.
// The function shall be public
//#]
void OMTimeoutPool::removeTm(const OMTimeout* const tm) {
    //#[ operation removeTm(OMTimeout)
    (void)OMOSSpecific::lock(tmPoolGuard);
    // Iterate down the busy list looking for the requested timeout. 
    OMTimeout* aTimeout = itsBusyList;
    OMTimeout* parent = NULL;
    OMTimeout* next = NULL;
    
    while ( aTimeout != NULL ) {  
    	next = aTimeout->getItsNextTimeout();
    	if (aTimeout == tm) 
    	{
    		// found it, so remove it from the busy list
    		aTimeout->setDestination(NULL);
    	 	if ( parent == NULL ) {
    	  		itsBusyList = next;
    		} 
    		else 
    		{
    			parent->setItsNextTimeout( next );
    		}
    		// Add the tm to itsFreeList
    		aTimeout->setItsNextTimeout( itsFreeList );
    		itsFreeList = aTimeout; 
    		break;
    	} 
    	else 
    	{
    		/*LDRA_INSPECTED 70 D : MISRA-C++:2008 0-1-6,0-1-9: "parent" is used in the loop*/
    		parent = aTimeout;
    	}
    	aTimeout = next;
    }
    
    (void)OMOSSpecific::unlock(tmPoolGuard);
    
    //#]
}

//#[ requirement(s)
// Realizes requirement Set Timeout FreeList #SXF_LR.1.8.7:
// The function has one parameter of 'OMTimeout* const' type.
// Function shall set its 'itsFreeList' attribute to the value of this parameter (even, if the parameter value is NULL).
// The function shall be public.
//#]
/*LDRA_INSPECTED 62 D 1 : MISRA-C++:2008 7-1-2: "p_OMTimeout" parameter cannot be const */
void OMTimeoutPool::setItsFreeList(OMTimeout* const p_OMTimeout) {
    //#[ operation setItsFreeList(OMTimeout)
    itsFreeList = p_OMTimeout;
    //#]
}

/*********************************************************************
	File Path	: ../OMTimeoutPool.cpp
*********************************************************************/
